#
# Virtual Traces Experiment Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2008.
# See PATCH_NOTE.
#


import math, udg

TICK = "system.cpu.fetch: Running stage."
COMMIT = "system.cpu.commit: Committed pc"
[ TYPE_VT, TYPE_BB ] = "VB"

class Edge_Item(udg.Traceable_Node):
    def __init__(self, target, label):
        self.target = [ target ]
        self.text = label

    def successors(self):
        return self.target

    def label(self):
        return self.text

class Stats_Obj(udg.Traceable_Node):
    def __init__(self, pc):
        udg.Traceable_Node.__init__(self)
        self.start = pc
        self.timings = dict()

    def Transition(self, next_vt, ticks, bb_count):
        key = (next_vt, bb_count, ticks)
        try:
            self.timings[ key ] += 1
        except KeyError:
            self.timings[ key ] = 1

    def Get_Stats(self):
        if ( len(self.timings) == 0 ):
            return None

        total_value = 0
        total_count = 0

        timings_list = self.timings.items()
        timings_list.sort()
        for ((next_vt, bb_count, ticks), count) in timings_list:
            total_count += count
            total_value += ticks * count

        assert total_count > 0
        mean = float(total_value) / float(total_count)
        total_variance = 0.0

        for ((next_vt, bb_count, ticks), count) in timings_list:
            v = float(ticks) - mean
            total_variance += v * v * float(count)

        variance = total_variance / float(total_count)
        std_dev = math.sqrt(variance)

        return (total_count, mean, std_dev)

    def Get_Stats_Text(self):
        rc = self.Get_Stats()
        if ( rc == None ):
            return None

        (total_count, mean, std_dev) = rc
        return '0x%010x, %5u, %5.2f, %5.2f' % (
                    self.start, total_count, mean, std_dev)

    def Get_Points(self):
        return '0x%010x, %s' % (self.start, ', '.join(
            [ '%x(%u)' % (ticks, count) 
                for ((next_vt, bb_count, ticks), count) in self.timings.iteritems() ]))

    def successors(self):
        return [ Edge_Item(next_vt, '%x(%u) - %u' % (ticks, count, bb_count))
            for ((next_vt, bb_count, ticks), count) 
                in self.timings.iteritems() ]

    def label(self):
        return '0x%x' % self.start


def Load_Stats(meta_path_in, commit_out):
    meta_inst_tbl = dict()
    meta_path = []
    meta_inst_list = []

    for line in file(meta_path_in):
        fields = line.split()
        if ( len(fields) >= 2 ):
            pc = int(fields[ 1 ], 16)

            if ( fields[ 0 ] == 'VTS' ):
                # Virtual Trace Start
                if ( not meta_inst_tbl.has_key(pc) ):
                    # Create new VT
                    meta_inst = Stats_Obj(pc)
                    meta_inst_tbl[ pc ] = meta_inst
                    meta_inst_list.append(meta_inst)
                else:
                    # Reuse existing VT
                    meta_inst = meta_inst_tbl[ pc ]

                meta_path.append((TYPE_VT, pc))

            elif ( fields[ 0 ] == 'BBS' ):
                # Basic block start
                meta_path.append((TYPE_BB, pc))
            
    meta_path.reverse()

    bb_count = 0
    start_tick = final_tick = ticks = 0
    vt = None
    for line in file(commit_out):
        x = line.find(':')
        if ( x < 0 ):
            continue

        x += 2
        line = line[ x: ]
        if ( line.startswith(TICK) ):
            ticks += 1

        elif ( line.startswith(COMMIT) ):
            pc = int(line[ len(COMMIT): ].split()[ 0 ], 16)

            max_records_left = 2

            while (( len(meta_path) != 0 )
            and ( max_records_left >= 0 )):

                (record_type, cmp_pc) = meta_path[ -1 ]

                #print 'top %s %x  c %x  r %u' % (record_type, cmp_pc, pc,
                #            len(meta_path))

                if ( cmp_pc == pc ):
                    if ( record_type == TYPE_VT ):
                        # New trace
                        next_vt = meta_inst_tbl[ pc ]
                        if ( vt != None ):
                            vt.Transition(next_vt, ticks - start_tick,
                                    bb_count)
                        else:
                            ticks = 0

                        vt = next_vt
                        start_tick = ticks
                        final_tick = ticks
                        bb_count = 0

                    elif ( record_type == TYPE_BB ):
                        # New basic block
                        bb_count += 1

                    # Advance to next record
                    meta_path.pop()
                    max_records_left -= 1
                else:
                    max_records_left = -1

    return (meta_inst_list, final_tick)

def VT_Stats(name, out_label, stats_source):
    (stats, ticks) = Load_Stats("%s_path.txt" % name, stats_source)
    udg.Trace_UDG_File("%s_%s.udg" % (name, out_label), stats[ : 1 ])
    out = file("%s_%s.txt" % (name, out_label), 'wt')
    for vt in stats:
        rc = vt.Get_Points()
        if ( rc != None ):
            out.write(rc)
            out.write('\n')
    out.close()
    return ticks

def Main():
    print 'VT mode', VT_Stats("ndes", "vtstats", "ndes_run_01")
    print 'Normal mode', VT_Stats("ndes", "nstats", "ndes_normal.txt")

